<p class="logo"><a href="http://doc.openerp.com/">
  <img class="logo" src="{{ pathto('_static/openerp.png', 1) }}" alt="Logo"/>
</a></p>

<h3>Other Docs</h3>
<ul>
  <li><a href="http://doc.openerp.com/trunk/developers">OpenERP Developers Documentation</a></li>
  <li><a href="http://doc.openerp.com/trunk/developers/web">OpenERP Web Developers Documentation</a></li>
  <li><a href="http://doc.openerp.com/trunk/users">OpenERP Users Documentation</a></li>
</ul>

<h3>Useful Links</h3>
<ul>
  <li><a href="http://www.openerp.com/">The OpenERP website</a></li>
  <li><a href="http://python.org/">The Python programming language</a></li>
</ul>
