# -*- coding: utf-8 -*-
##############################################################################
#    
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2004-2009 Tiny SPRL (<http://tiny.be>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.     
#
##############################################################################

from rml2txt import parseString, parseNode

#.apidoc title: RML to TXT engine

""" This engine is the minimalistic renderer of RML documents into text files,
    using spaces and newlines to format.

    It was needed in some special applications, where legal reports need to be
    printed in special (dot-matrix) printers.
"""

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:

