import test_base
import test_expression
import test_ir_attachment
import test_ir_values
import test_menu
import test_search
import test_views

checks = [
    test_base,
    test_expression,
    test_ir_attachment,
    test_ir_values,
    test_menu,
    test_search,
    test_views,
]
