# -*- coding: utf-8 -*-
##############################################################################
#    
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2004-2009 Tiny SPRL (<http://tiny.be>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.     
#
##############################################################################

pageSize = {
    'A4': (210,297),
    'A5': (148.5,105)
}

odt_namespace = {
    "office":"{urn:oasis:names:tc:opendocument:xmlns:office:1.0}",
    "style":"{urn:oasis:names:tc:opendocument:xmlns:style:1.0}",
    "text":"{urn:oasis:names:tc:opendocument:xmlns:text:1.0}",
    "table":"{urn:oasis:names:tc:opendocument:xmlns:table:1.0}",
    "draw":"{urn:oasis:names:tc:opendocument:xmlns:drawing:1.0}",
    "fo":"{urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0}",
    "xlink":"{http://www.w3.org/1999/xlink}",
    "dc":"{http://purl.org/dc/elements/1.1/}",
    "meta":"{urn:oasis:names:tc:opendocument:xmlns:meta:1.0}",
    "number":"{urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0}",
    "svg":"{urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0}",
    "chart":"{urn:oasis:names:tc:opendocument:xmlns:chart:1.0}",
    "dr3d":"{urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0}",
    "math":"{http://www.w3.org/1998/Math/MathML}",
    "form":"{urn:oasis:names:tc:opendocument:xmlns:form:1.0}",
    "script":"{urn:oasis:names:tc:opendocument:xmlns:script:1.0}",
    "ooo":"{http://openoffice.org/2004/office}",
    "ooow":"{http://openoffice.org/2004/writer}",
    "oooc":"{http://openoffice.org/2004/calc}",
    "dom":"{http://www.w3.org/2001/xml-events}" }

sxw_namespace = {
    "office":"{http://openoffice.org/2000/office}",
    "style":"{http://openoffice.org/2000/style}",
    "text":"{http://openoffice.org/2000/text}",
    "table":"{http://openoffice.org/2000/table}",
    "draw":"{http://openoffice.org/2000/drawing}",
    "fo":"{http://www.w3.org/1999/XSL/Format}",
    "xlink":"{http://www.w3.org/1999/xlink}",
    "dc":"{http://purl.org/dc/elements/1.1/}",
    "meta":"{http://openoffice.org/2000/meta}",
    "number":"{http://openoffice.org/2000/datastyle}",
    "svg":"{http://www.w3.org/2000/svg}",
    "chart":"{http://openoffice.org/2000/chart}",
    "dr3d":"{http://openoffice.org/2000/dr3d}",
    "math":"{http://www.w3.org/1998/Math/MathML}",
    "form":"{http://openoffice.org/2000/form}",
    "script":"{http://openoffice.org/2000/script}",
    "ooo":"{http://openoffice.org/2004/office}",
    "ooow":"{http://openoffice.org/2004/writer}",
    "oooc":"{http://openoffice.org/2004/calc}",
    "dom":"{http://www.w3.org/2001/xml-events}"}
# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:

